"
 COPYRIGHT (c) 1989 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

View subclass:#Scroller
       instanceVariableNames:'thumbOrigin thumbHeight thumbColor thumbFrameColor
                              scrollAction moveDirection
                              thumbFrame thumbLevel
                              scrolling pressOffset
                              synchronousOperation
                              shadowForm lightForm inset
                              thumbShadowColor thumbLightColor
                              thumbSoftEdge
                              thumbHalfShadowColor thumbHalfLightColor
                              thumbFrameSizeDifference
                              tallyLevel tallyMarks'
       classVariableNames:   'HandleShadowForm HandleLightForm'
       poolDictionaries:''
       category:'Views-Interactors'
!

Scroller comment:'

COPYRIGHT (c) 1989 by Claus Gittinger
              All Rights Reserved

$Header: /cvs/stx/stx/libwidg/Scroller.st,v 1.6 1994-01-13 00:18:10 claus Exp $

written spring/summer 89 by claus
'!

!Scroller class methodsFor:'documentation'!

documentation
"
this class implements the scroller for scrollbars.
it can also be used by itself for scrollbars without step-buttons.
When moved, a predefined action is performed.
Beside the obvious 3D rectangle, a scroller may draw a know-form
(as in NeXT) or little tally marks (as on SGI) in itself.
These are controlled by the shadowForm, lightForm, tallyLevel and tallyMarks
instance variables.

Instance variables:

thumbOrigin                 <Number>        origin of thumb (in percent)
thumbHeight                 <Number>        height of thumb (in percent)
thumbColor                  <Color>         color of thumb
thumbFrameColor             <Color>         color of the frame around the thumb
scrollAction                <Block>         1 arg block to be evaluated when scrolled
                                            (arg is position in percent)
moveDirection               <Symbol>        #x or #y
thumbFrame                  <Rectangle>     frame of thumb in pixels (cached)
thumbLevel                  <Number>        level of thumb if 3d
scrolling                   <Boolean>       true during scroll
pressOffset                 <Number>        temporary (offset into frame when move started)
synchronousOperation        <Boolean>       true if synchronous (i.e. dont wait till release
                                            to perform action)
shadowForm                  <Form>          bitmap of knob if any (shadow part)
lightForm                   <Form>          bitmap of knob if any (light part)
inset                       <Integer>       number of pixels to inset thumb from view borders
thumbShadowColor            <Color>         color do draw dark parts of thumb
thumblightColor             <Color>         color to draw light parts of thumb
thumbSoftEdge               <Boolean>       true if edges of thumb are to appear smooth
thumbHalfShadowColor        <Color>         used to draw smooth edges
thumbHalfLightColor         <Color>         used to draw smooth edges
thumbFrameSizeDifference    <Integer>       number of pixels the thumb is larger than 
                                            it should be
tallyLevel                  <Integer>       if not zero, specifies if tally-marks should
                                            go into or out of the display (actually only <0/>0 is checked)
                                            I dont know of a better word for these ...
tallyMarks                  <Integer>       number of tally marks
"
! !

!Scroller class methodsFor:'defaults'!

handleShadowFormOn:aDisplay
    "answer the form used for the handles shadow area;
     cache the one for Display for the next round"

    |f|

    ((aDisplay == Display) and:[HandleShadowForm notNil]) ifTrue:[
        ^ HandleShadowForm
    ].
    f := Form fromFile:'HandleShadow.xbm' resolution:100 on:aDisplay.
    f isNil ifTrue:[
        f := Form width:8 height:8 fromArray:#(2r00111100
                                               2r01100000
                                               2r11000000
                                               2r11000000
                                               2r11000000
                                               2r11000000
                                               2r01000000
                                               2r00000000)
                                           on:aDisplay
    ].
    (aDisplay == Display) ifTrue:[
        HandleShadowForm := f
    ].
    ^ f
!

handleLightFormOn:aDisplay
    "answer the form used for the handles light area;
     cache the one for Display for the next round"

    |f|

    ((aDisplay == Display) and:[HandleLightForm notNil]) ifTrue:[
        ^ HandleLightForm
    ].
    f := Form fromFile:'HandleLight.xbm' resolution:100 on:aDisplay.
    f isNil ifTrue:[
        f := Form width:8 height:8 fromArray:#(2r00000000
                                               2r00000010
                                               2r00000011
                                               2r00000011
                                               2r00000011
                                               2r00000011
                                               2r00000110
                                               2r00111100)
                                              on:aDisplay
    ].
    (aDisplay == Display) ifTrue:[
        HandleLightForm := f
    ].
    ^ f
! !

!Scroller methodsFor:'initialization'!

initialize
    "initialize - setup instvars from defaults"

    super initialize.
    moveDirection := #y.

    scrolling := false.
    synchronousOperation := true.

    thumbOrigin := 0.
    thumbHeight := 100.

    inset := 1.

    self computeThumbFrame
!

initStyle
    "initialize style dep. stuff"

    super initStyle.

    viewBackground := Color grey.
    thumbLevel := 0.
    thumbSoftEdge := false.
    tallyLevel := 0.
    tallyMarks := 0.

    self is3D ifTrue:[
        thumbSoftEdge := false.
        inset := 0.

        style == #next ifTrue:[
            self level:0.
            self borderWidth:1.
            inset := 1.
            thumbSoftEdge := true.
            thumbLevel := 2.
            thumbColor := Color lightGrey
        ] ifFalse:[
            style == #motif ifTrue:[
                self level:0.
                inset := 1.
                thumbLevel := 2.
                viewBackground := Color lightGrey.
                thumbColor := viewBackground.
            ] ifFalse:[
                style == #iris ifTrue:[
                    self level:-1.
                    thumbLevel := 3.
                    thumbSoftEdge := true.
                    thumbColor := viewBackground.
                    tallyLevel := 1.
                    tallyMarks := 3.
                ] ifFalse:[
                    style == #mswindows ifTrue:[
                        self level:0.
                        self borderWidth:1.
                        thumbLevel := 2.
                        thumbColor := Color lightGrey.
                        thumbSoftEdge := true.
                        viewBackground := Color grey:80.
                    ] ifFalse:[
                        self level:-1.
                        thumbLevel := 2.
                        thumbColor := Color lightGrey
                    ]
                ]
            ]
        ].
        device hasGreyscales ifFalse:[
            thumbColor := Color grey
        ].
    ] ifFalse:[
        thumbColor := White.
        inset := 1
    ].

    thumbShadowColor := shadowColor.
    thumbLightColor := lightColor.
    thumbSoftEdge ifTrue:[
        device hasGreyscales ifTrue:[
            thumbHalfShadowColor := halfShadowColor.
            thumbHalfLightColor := halfLightColor
        ] ifFalse:[
            thumbHalfShadowColor := Color darkGrey "halfShadowColor".
            thumbHalfLightColor := White
        ]
    ].
    device hasGreyscales ifFalse:[
        thumbShadowColor := Black.
        thumbLightColor := White.
        viewBackground := Color veryLightGrey "White"
    ].

    thumbFrameColor := Black.

    style == #next ifTrue:[
        shadowForm := self class handleShadowFormOn:device.
        lightForm := self class handleLightFormOn:device
    ] ifFalse:[
        shadowForm := lightForm := nil
    ].

    drawableId notNil ifTrue:[
        self computeThumbFrame
    ]
!

initCursor
    "set the cursor - a hand"

    cursor := Cursor hand
!

initEvents
    self enableButtonEvents.
    self enableButtonMotionEvents
! !

!Scroller methodsFor:'accessing'!

is3D
    style == #mswindows ifTrue:[^ true].
    ^ super is3D
!

asynchronousOperation
    "set scroll-mode to be asynchronous - scroll action is performed after
     scrolling, when mouse-button is finally released"
     
    synchronousOperation := false
!

synchronousOperation
    "set scroll-mode to be synchronous - scroll action is performed for 
     every movement of thumb"
     
    synchronousOperation := true
!

scrollAction:aBlock
    "set the scroll action, aBlock which is evaluated when scrolled"

    scrollAction := aBlock
!

scrollAction
    "answer the scroll action block"

    ^ scrollAction
!

thumbOrigin
    "answer the thumbs origin (in percent)"

    ^ thumbOrigin
!

thumbOrigin:newOrigin
    "set the thumbs origin (in percent)"

    |realNewOrigin oldFrame oldTop oldBot thumbTop thumbBot
     tH "{ Class: SmallInteger }"
     tW delta left|

    ((newOrigin + thumbHeight) > 100) ifTrue:[
        realNewOrigin := 100 - thumbHeight
    ] ifFalse: [
        realNewOrigin := newOrigin
    ].
    (realNewOrigin > 100) ifTrue:[
        realNewOrigin := 100
    ] ifFalse: [
        (realNewOrigin < 0) ifTrue:[
            realNewOrigin := 0
        ]
    ].
    (realNewOrigin = thumbOrigin) ifFalse:[
        oldFrame := thumbFrame.
        thumbOrigin := realNewOrigin.
        self computeThumbFrame.
        (thumbHeight = 100) ifTrue:[^ self].

        shown ifTrue:[
            (thumbFrame ~~ oldFrame) ifTrue:[
                tH := thumbFrame height.
                tW := thumbFrame width.
                oldTop := oldFrame top.
                oldBot := oldTop + tH.

                thumbTop := thumbFrame top.
                thumbBot := thumbTop + tH.

                left := thumbFrame left.

                (oldBot >= height) ifTrue:[
                    "cannot copy - thumb was below end"
                    self drawThumbBackgroundInX:left y:oldTop
                                          width:tW height:(height - oldTop).
                    self drawThumb.
                    ^ self
                ].

                self copyFrom:self x:left y:oldTop
                                 toX:left y:thumbTop
                               width:tW height:tH.

                self catchExpose.

                oldTop > thumbTop ifTrue:[
                    delta := oldTop - thumbTop.
                    oldTop > thumbBot ifTrue:[
                        self drawThumbBackgroundInX:left y:oldTop
                                              width:tW height:(tH + 1)
                    ] ifFalse:[
                        self drawThumbBackgroundInX:left y:thumbBot 
                                              width:tW height:delta
                    ]
                ] ifFalse:[
                    delta := thumbTop - oldTop.
                    oldBot < thumbTop ifTrue:[
                        self drawThumbBackgroundInX:left y:oldTop
                                              width:tW height:(tH + 1)
                    ] ifFalse:[
                        self drawThumbBackgroundInX:left y:oldTop 
                                              width:tW height:delta
                    ]
                ].
                self waitForExpose
            ]
        ]
    ]
!

thumbHeight
    "answer the thumbs height (in percent)"

    ^ thumbHeight
!

thumbHeight:newHeight
    "set the thumbs height (in percent)"

    |realNewHeight oldFrame|

    (newHeight > 100) ifTrue:[
        realNewHeight := 100
    ] ifFalse:[
        realNewHeight := newHeight
    ].
    (realNewHeight = thumbHeight) ifFalse:[
        oldFrame := thumbFrame.
        thumbHeight := realNewHeight.
        self computeThumbFrame.
        shown ifTrue:[
            (oldFrame ~~ thumbFrame) ifTrue:[
                self drawThumbBackgroundInX:(oldFrame left)
                                          y:(oldFrame top) 
                                      width:(oldFrame width) 
                                     height:(oldFrame height).
                self drawThumb
            ]
        ]
    ]
!

thumbOrigin:newOrigin thumbHeight:newHeight
    "set both thumbs height and origin (in percent)"

    |realNewOrigin realNewHeight old new same|

    (newHeight > 100) ifTrue:[
        realNewHeight := 100
    ] ifFalse:[
        realNewHeight := newHeight
    ].
    ((newOrigin + realNewHeight) > 100) ifTrue:[
        realNewOrigin := 100 - realNewHeight
    ] ifFalse: [
        realNewOrigin := newOrigin
    ].
    (realNewOrigin < 0) ifTrue: [
        realNewOrigin := 0
    ].

    same := (realNewHeight = thumbHeight).
    same ifTrue:[
        same := (realNewOrigin = thumbOrigin)
    ].

    same ifFalse:[
        old := self absFromPercent:thumbOrigin.
        new := self absFromPercent:realNewOrigin.
        (old == new) ifTrue:[
            old := self absFromPercent:thumbHeight.
            new := self absFromPercent:realNewHeight.
            (old == new) ifTrue:[^ self]
        ].

        shown ifTrue:[
            self drawThumbBackgroundInX:(thumbFrame left)
                                      y:(thumbFrame top) 
                                  width:(thumbFrame width) 
                                 height:(thumbFrame height).
        ].
        thumbOrigin := realNewOrigin.
        thumbHeight := realNewHeight.
        self computeThumbFrame.
        shown ifTrue:[
            self drawThumb
        ]
    ]
!

setThumbFor:aView
    "get contents and size info from aView and adjust thumb"

    |percentSize percentOrigin totalHeight|

    aView isNil ifTrue:[
        totalHeight := 0
    ] ifFalse:[
        totalHeight := aView heightOfContents
    ].
    (totalHeight = 0) ifTrue:[
        percentSize := 100.
        percentOrigin := 100
    ] ifFalse:[
        percentSize := (aView innerHeight) * 100.0 / totalHeight.
        percentOrigin := (aView yOriginOfContents) * 100.0 / totalHeight.
        percentOrigin + percentSize > 100.0 ifTrue:[
            "actually showing stuff below contents of view"
"
            totalHeight := aView yOriginOfContents + aView innerHeight.
            percentSize := (aView innerHeight) * 100.0 / totalHeight.
            percentOrigin := (aView yOriginOfContents) * 100.0 / totalHeight
"
        ]
    ].
    (percentSize = thumbHeight) ifTrue:[
        self thumbOrigin:percentOrigin
    ] ifFalse:[
        (percentOrigin = thumbOrigin) ifTrue:[
            self thumbHeight:percentSize
        ] ifFalse:[
            self thumbOrigin:percentOrigin thumbHeight:percentSize
        ]
    ]
!

setThumbHeightFor:aView
    "get contents and size info from aView and adjust thumb height"

    |percent totalHeight|

    totalHeight := aView heightOfContents.
    (totalHeight = 0) ifTrue:[
        percent := 100
    ] ifFalse:[
        percent := (aView innerHeight) * 100.0 / totalHeight
    ].
    self thumbHeight:percent
!

setThumbOriginFor:aView
    "get contents and size info from aView and adjust thumb origin"

    |percent totalHeight|

    totalHeight := aView heightOfContents.
    (totalHeight = 0) ifTrue:[
        percent := 100
    ] ifFalse:[
        percent := (aView yOriginOfContents) * 100.0 / totalHeight
    ].
    self thumbOrigin:percent
!

thumbColor:aColor
    "change the color of the thumb"

    thumbColor := aColor on:device.
    self is3D ifTrue:[
        thumbShadowColor := aColor darkened on:device.
        thumbLightColor := aColor lightened on:device.
        thumbHalfShadowColor := thumbShadowColor darkened on:device.
        thumbHalfLightColor := thumbLightColor lightened on:device.
    ].
    shown ifTrue:[
        self redraw
    ]
!

thumbFrame
    "return the area used by the thumbFrame (in device coordinates).
     Allows access to the thumbs physical screen position, for
     example to position a label below (see Slider-Examples)"

    ^ thumbFrame
! !

!Scroller methodsFor:'private'!

absFromPercent:percent
    "given a percentage, compute number of pixels"

    |fullSize|

    (moveDirection == #y) ifTrue:[
        fullSize := height
    ] ifFalse:[
        fullSize := width
    ].
    ^ ((percent * (fullSize - (margin * 2))) / 100) rounded
!

percentFromAbs:absValue
    "given a number of pixels, compute percentage"

    |fullSize val|

    (moveDirection == #y) ifTrue:[
        fullSize := height
    ] ifFalse:[
        fullSize := width
    ].

    val := absValue / (fullSize - thumbFrameSizeDifference - (margin * 2)) * 100.
    val < 0 ifTrue:[^ 0].
    val > 100 ifTrue:[^ 100].
    ^ val
!

computeThumbFrame
    "compute the thumbs frame (a rectangle) whenever thumb is moved, changed
     height or the scrollers size has changed"

    |np1 np2 ns1 ns2 nh nw ny nx t sz1 sz2|

    np1 := (self absFromPercent:thumbOrigin) + margin.
    ns1 := self absFromPercent:thumbHeight.
    thumbFrameSizeDifference := 0.
    (moveDirection == #y) ifTrue:[
        sz1 := height.
        sz2 := width
    ] ifFalse:[
        sz1 := width.
        sz2 := height
    ].
    self is3D ifTrue:[
        np2 := margin + inset.     
        ns2 := sz2 - (margin * 2) - (inset * 2).
        "
         do not make thumb too small (for handle)
        "
        (ns1 < (10 + (2 * thumbLevel))) ifTrue:[
            t := ns1.
            ns1 := 10 + (2 * thumbLevel).
            thumbFrameSizeDifference := ns1 - t
        ]
    ] ifFalse:[
        np2 := inset.
        ns2 := sz2 - (inset * 2).

        "
         do not make thumb too small (uncatchable)
        "
        (ns1 < 4) ifTrue:[
            t := ns1.
            ns1 := 4.
            thumbFrameSizeDifference := ns1 - t
        ]
    ].
    "
     oops - if height has been increased, we have to adjust
     the origin
    "
    (thumbFrameSizeDifference == 0) ifFalse:[
        np1 := ((thumbOrigin * (sz1 - thumbFrameSizeDifference - (margin * 2))) / 100) rounded + margin
    ].

    (moveDirection == #y) ifTrue:[
        ny := np1.
        nx := np2.
        nh := ns1.
        nw := ns2.
        ny + nh + margin > height ifTrue:[
            ny := height - margin - nh
        ]
    ] ifFalse:[
        nx := np1.
        ny := np2.
        nw := ns1.
        nh := ns2.
        nx + nw + margin > width ifTrue:[
            nx := width - margin - nw
        ]
    ].

    "
     do not create new Rectangle if its the same anyway
    "
    thumbFrame notNil ifTrue:[
        (ny == thumbFrame top) ifTrue:[
          (nx == thumbFrame left) ifTrue:[
            (nh == thumbFrame height) ifTrue:[
              (nw == thumbFrame width) ifTrue:[ ^ self]
            ]
          ]
        ]
    ].
    thumbFrame := Rectangle left:nx top:ny width:nw height:nh
! !

!Scroller methodsFor:'drawing'!

drawHandleFormAtX:x y:y
    |oldFg oldBg oldFun|

    oldFg := foreground.
    oldBg := background.
    oldFun := function.

    thumbShadowColor := thumbShadowColor on:device.
    thumbLightColor := thumbLightColor on:device.
    thumbColor := thumbColor on:device.

    "kludge for now"
    ((thumbShadowColor colorId notNil)
     and:[(thumbLightColor colorId notNil)
         and:[thumbColor colorId notNil]])
     ifTrue:[
        self foreground:thumbColor background:(Color noColor) function:#xor.
        self displayOpaqueForm:shadowForm x:x y:y.
        self foreground:thumbShadowColor function:#or.
        self displayOpaqueForm:shadowForm x:x y:y.
        self foreground:thumbColor function:#xor.
        self displayOpaqueForm:lightForm x:x y:y.
        self foreground:thumbLightColor function:#or.
        self displayOpaqueForm:lightForm x:x y:y.
        self foreground:oldFg background:oldBg function:#copy
    ] ifFalse:[
false ifTrue:[
        "turn off shadow bits"
        self foreground:(Color noColor) background:(Color allColor) function:#and.

        self displayOpaqueForm:shadowForm x:x y:y.
        "turn off light bits"
        self displayOpaqueForm:lightForm x:x y:y.

        "draw shadow bits"
        self paint:shadowColor.
        self function:#or.
        self displayOpaqueForm:shadowForm x:x y:y.

        "draw light bits"
        self paint:lightColor.
        self displayOpaqueForm:lightForm x:x y:y.
        self foreground:oldFg background:oldBg function:oldFun
]
    ]
!

drawThumbBackgroundInX:x y:y width:w height:h
    "draw part of the thumbs background; defined as a separate
     method, to allow drawing of arbitrary patterns under thumb."

    self clearRectangleX:x y:y width:w height:h.
!

drawThumb
    "draw the thumb"

    |handleX handleY l t 
     w "{ Class: SmallInteger }"
     h "{ Class: SmallInteger }"
     x "{ Class: SmallInteger }"
     y "{ Class: SmallInteger }"
     mm xL xR yT yB color1 color2 savEdge|

    ((thumbHeight >= 100) or:[thumbFrame height >= height]) ifTrue:[^ self].

    l := thumbFrame left.
    t := thumbFrame top.
    w := thumbFrame width.
    h := thumbFrame height.
    self paint:thumbColor.
    self fillRectangleX:l y:t width:w height:h.

    self is3D ifFalse:[
        self paint:thumbFrameColor.
        self displayRectangle:thumbFrame.
        ^ self
    ].

    "what a kludge - must be a parameter to drawEdge..."
    savEdge := softEdge.
    softEdge := thumbSoftEdge.
    self drawEdgesForX:l y:t width:w height:h level:thumbLevel
                shadow:thumbShadowColor light:thumbLightColor
                halfShadow:thumbHalfShadowColor halfLight:thumbHalfLightColor.
    softEdge := savEdge.

    (tallyLevel == 0 or:[tallyMarks == 0]) ifTrue:[
        shadowForm notNil ifTrue:[
            handleX := l + ((w - 8) // 2).
            handleY := t + ((h - 8) // 2).
            self drawHandleFormAtX:handleX y:handleY
        ].
        ^ self
    ].

    "iris style - draw tallys"

    tallyLevel > 0 ifTrue:[
        color1 := thumbLightColor.
        color2 := thumbShadowColor.
    ] ifFalse:[
        color1 := thumbShadowColor.
        color2 := thumbLightColor.
    ].

    "draw tally marks"

    (moveDirection == #y) ifTrue:[
        self paint:color1.
        y := t + (h // 2) - 1.
        xL := l + thumbLevel - 1.
        xR := l + w - thumbLevel "+ 1".
        self displayLineFromX:xL y:y toX:xR y:y.
        y := y + 1.
        self paint:color2.
        self displayLineFromX:xL y:y toX:xR y:y.

        tallyMarks > 1 ifTrue:[
            "dont draw other marks if there is not enough space"

            mm := device verticalPixelPerMillimeter rounded.
            h > (mm * (tallyMarks * 2)) ifTrue:[
                y := y - 1 - mm.
                self paint:color1.
                self displayLineFromX:xL y:y toX:xR y:y.
                y := y + 1.
                self paint:color2.
                self displayLineFromX:xL y:y toX:xR y:y.

                y := y - 1 + mm + mm.
                self paint:color1.
                self displayLineFromX:xL y:y toX:xR y:y.
                y := y + 1.
                self paint:color2.
                self displayLineFromX:xL y:y toX:xR y:y
            ]
        ]
    ] ifFalse:[
        x := l + (w // 2) - 1.
        yT := t + thumbLevel - 1.
        yB := t + h - thumbLevel "+ 1".
        self paint:color1.
        self displayLineFromX:x y:yT toX:x y:yB.
        self paint:color2.
        x := x + 1.
        self displayLineFromX:x y:yT toX:x y:yB.

        tallyMarks > 1 ifTrue:[
            "dont draw other marks if there is not enough space"

            mm := device horizontalPixelPerMillimeter rounded.
            w > (mm * (tallyMarks * 2)) ifTrue:[
                x := x - 1 - mm.
                self paint:color1.
                self displayLineFromX:x y:yT toX:x y:yB.
                x := x + 1.
                self paint:color2.
                self displayLineFromX:x y:yT toX:x y:yB.

                x := x - 1 + mm + mm.
                self paint:color1.
                self displayLineFromX:x y:yT toX:x y:yB.
                x := x + 1.
                self paint:color2.
                self displayLineFromX:x y:yT toX:x y:yB
            ]
        ]
    ]
! !

!Scroller methodsFor:'events'!

redrawX:x y:y width:w height:h
    (y > thumbFrame bottom) ifTrue:[
        self drawThumbBackgroundInX:x y:y width:w height:h.
        ^ self
    ].
    ((y + h) < thumbFrame top) ifTrue:[
        self drawThumbBackgroundInX:x y:y width:w height:h.
        ^ self
    ].
    self drawThumbBackgroundInX:0 y:0 width:width height:height.
    self drawThumb
!

redraw
    "redraw"

    shown ifTrue:[
        self drawThumbBackgroundInX:0 y:0 width:width height:height.
        self drawThumb
    ]
!

sizeChanged:how
    "size of scroller changed - recompute thumbs frame and redraw it"

    self computeThumbFrame.
    self redraw
!

buttonPress:button x:x y:y
    "button was pressed - if above thumb, page up; if below thumb, page down;
     otherwise start scrolling"

    |curr limit1 limit2|

    (moveDirection == #y) ifTrue:[
        curr := y.
        limit1 := thumbFrame top.
        limit2 := thumbFrame bottom
    ] ifFalse:[
        curr := x.
        limit1 := thumbFrame left.
        limit2 := thumbFrame right
    ].

    (curr < limit1) ifTrue:[
        "page up/left"
        self thumbOrigin:(thumbOrigin - thumbHeight).
        scrollAction notNil ifTrue:[
            scrollAction value:thumbOrigin
        ]
    ] ifFalse:[
        (curr > limit2) ifTrue:[
            "page down/right"
            self thumbOrigin:(thumbOrigin + thumbHeight).
            scrollAction notNil ifTrue:[
                scrollAction value:thumbOrigin
            ]
        ] ifFalse:[
            pressOffset := curr - limit1.
            scrolling := true
        ]
    ]
!

buttonMultiPress:button x:x y:y
    ^ self buttonPress:button x:x y:y
!

buttonShiftPress:button x:x y:y
    "mouse-click with shift - jump to position"

    |pos newThumbOrigin curr limit org|

    (moveDirection == #y) ifTrue:[
        curr := y.
        limit := height.
        org := thumbFrame top
    ] ifFalse:[
        curr := x.
        limit := width.
        org := thumbFrame left
    ].

    (curr < 0) ifTrue:[                        "check against limits"
        pos := 0
    ] ifFalse:[
        (curr > limit) ifTrue:[
            pos := limit
        ] ifFalse:[
            pos := curr
        ]
    ].

    newThumbOrigin := self percentFromAbs:pos.
    self thumbOrigin:newThumbOrigin.
    scrollAction notNil ifTrue:[
        scrollAction value:thumbOrigin
    ].
    pressOffset := curr - org.
    scrolling := true
!

buttonMotion:button x:x y:y
    "mouse-button was moved while pressed;
     redraw thumb at its new position and, if scroll-mode is asynchronous, 
     the scroll action is performed"

    |pos newThumbOrigin curr limit|

    scrolling ifFalse: [ ^ self ].          "should not happen"

    (moveDirection == #y) ifTrue:[
        curr := y.
        limit := height
    ] ifFalse:[
        curr := x.
        limit := width
    ].

    (curr < 0) ifTrue:[                        "check against limits"
        pos := 0
    ] ifFalse:[
        (curr > limit) ifTrue:[
            pos := limit
        ] ifFalse:[
            pos := curr
        ]
    ].

    newThumbOrigin := self percentFromAbs:(pos - pressOffset).

    self thumbOrigin:newThumbOrigin.
    synchronousOperation ifTrue: [
        scrollAction notNil ifTrue:[
            scrollAction value:thumbOrigin
        ]
    ]
!

buttonRelease:button x:x y:y
    "mouse-button was released - if scroll-mode is asynchronous, the scroll
     action is now performed"

    scrolling ifTrue:[
        scrolling := false.
        synchronousOperation ifFalse: [
            scrollAction notNil ifTrue:[
                scrollAction value:thumbOrigin
            ]
        ]
    ]
! !
